#include "unit.h"
#include "model.h"
#include "gauge.h"
#include "sprite.h"
#include "square.h"
#include "morphing.h"
#include "menu.h"
#include "randgen.h"

#include <SDL2/SDL.h>
#include <iostream>
#include <string>
#include <map>
#include <stdlib.h>
#include <time.h>

#define TITLE "MiDriPod"
#define ICON_FILENAME "data\\midripod.bmp"
// #define ICON_FILENAME "data\\icon.bmp"
#define FULLSCREEN

#define HARDNESS_ROCK       10
#define HARDNESS_CRISTAL    30
#define HARDNESS_POWERGEM   15

void error(std::string msg) {
    std::cerr << msg << ": " << SDL_GetError() << std::endl;
    exit(EXIT_FAILURE);
}

void loop(SDL_Renderer * renderer);
void player_move(struct unit * cominex, struct square grid[GRID_WIDTH][GRID_HEIGHT]);
void build_level(struct square grid[GRID_WIDTH][GRID_HEIGHT]);

int main(int argc, char**argv) {
    SDL_Window * window;
    SDL_Renderer * renderer;
    SDL_Surface * icon;

    if(SDL_Init(SDL_INIT_VIDEO) != 0)
        error("Unable to init SDL");
    SDL_CreateWindowAndRenderer(0, 0, 0, &window, &renderer);
    if(window == NULL)
        error("Unable to open window");
    if(renderer == NULL)
        error("Unable to create renderer");

    SDL_SetWindowTitle(window, TITLE);
    SDL_SetWindowSize(window, WINDOW_WIDTH, WINDOW_HEIGHT);
    SDL_SetWindowPosition(window, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
    #ifdef FULLSCREEN
    SDL_SetWindowFullscreen(window, SDL_WINDOW_FULLSCREEN);
    #endif

    icon = SDL_LoadBMP(ICON_FILENAME);
    if(icon == NULL)
        error("Unable to open ICON_FILENAME file");
    SDL_SetWindowIcon(window, icon);

    srand(time(NULL));

    loop(renderer);

    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);
    SDL_FreeSurface(icon);
    SDL_Quit();

    return(EXIT_SUCCESS);
}

struct morphing * morphing_rock;
struct morphing * morphing_cristal;
struct morphing * morphing_powergem;
struct sprite * icon_energy;
struct sprite * icon_cristal;
struct sprite * icon_powergem;

void loop(SDL_Renderer * renderer) {
    SDL_Event event;
    struct square grid[GRID_WIDTH][GRID_HEIGHT];
    struct menu * building_menu;
    std::forward_list<struct unit *> units;
    std::map<std::string, struct model *> models;

    icon_energy   = new sprite(renderer, "data\\icon_energy.bmp");
    icon_cristal  = new sprite(renderer, "data\\icon_cristal.bmp");
    icon_powergem = new sprite(renderer, "data\\icon_powergem.bmp");
/*
    morphing_rock = new morphing(renderer, 10, {
                "data\\rock_soil.bmp",
                "data\\rock_cracked.bmp",
                "data\\rock.bmp"
                });
    morphing_cristal = new morphing(renderer, 30, {
                "data\\cristal_soil.bmp",
                "data\\cristal_very_cracked.bmp",
                "data\\cristal_cracked.bmp",
                "data\\cristal.bmp"
                });
*/
    morphing_rock = new morphing(renderer, HARDNESS_ROCK, "data\\rock.bmp");
    morphing_cristal = new morphing(renderer, HARDNESS_CRISTAL, "data\\gem_white.bmp");
    morphing_powergem = new morphing(renderer, HARDNESS_POWERGEM, "data\\gem_green.bmp");

    struct unit * midripod;
    struct unit * cominex;
    bool stop = false;

    models["midripod"] = new model("midripod", "data\\midripod.bmp", renderer);
    models["midripod"]->drill_power = 500;
    models["midripod"]->max_energy = 2;
    models["midripod"]->max_cristal = 15;
    models["midripod"]->max_powergem = 1;
    models["midripod"]->generate_energy = ENERGY_GENERATOR_THRESHOLD/100/3; // 1 each 5 sec.
    models["midripod"]->collect_cristal = true;
    models["midripod"]->collect_powergem = true;
    models["midripod"]->give_energy = true;

    models["cominex"] = new model("cominex", "data\\cominex.bmp", renderer);
    models["cominex"]->max_energy = 5;
    models["cominex"]->max_cristal = 5;
    models["cominex"]->max_powergem = 1;
    models["cominex"]->drill_power = 4;
    models["cominex"]->collect_cristal = true;
    models["cominex"]->collect_energy = true;
    models["cominex"]->give_cristal = true;
    models["cominex"]->give_powergem = true;
    // models["cominex"]->give_energy = true; // XXX
    models["cominex"]->speed = 1.0;

    models["energen"] = new model("energen", "data\\energen.bmp", renderer);
    models["energen"]->cristal_cost = 1;
    models["energen"]->max_energy = 5;
    models["energen"]->max_powergem = 3;
    models["energen"]->generate_energy = ENERGY_GENERATOR_THRESHOLD/100/3; // 1 each 3 sec.
    models["energen"]->give_energy = true;
    models["energen"]->collect_powergem = true;

    models["enerstash"] = new model("enerstash", "data\\enerstash.bmp", renderer);
    models["enerstash"]->cristal_cost = 1;
    models["enerstash"]->max_energy = 15;
    models["enerstash"]->collect_energy = true;
    models["enerstash"]->give_energy = true;

    models["midro"] = new model("midro", "data\\midro.bmp", renderer);
    models["midro"]->cristal_cost = 1;
    models["midro"]->max_energy = 3;
    models["midro"]->max_cristal = 3;
    models["midro"]->drill_power = 1;
    models["midro"]->give_cristal = true;
    models["midro"]->give_powergem = true;
    models["midro"]->collect_energy = true;
    models["midro"]->speed = 1.3;

    models["factory"] = new model("factory", "data\\factory.bmp", renderer);
    models["factory"]->cristal_cost = 3;
    models["factory"]->max_energy = 3;
    models["factory"]->max_cristal = 6;
    models["factory"]->max_powergem = 1;
    models["energen"]->generate_energy = ENERGY_GENERATOR_THRESHOLD/100/5; // 1 each 5 sec.
    models["factory"]->collect_energy = true;
    models["factory"]->collect_cristal = true;
    models["factory"]->collect_powergem = true;
    models["factory"]->autobuild = models["midro"];

    building_menu = new menu({models["energen"], models["enerstash"], models["factory"]});

    build_level(grid);

    midripod = new unit(models["midripod"], WINDOW_WIDTH/2, WINDOW_HEIGHT/2);
    cominex = new unit(models["cominex"], WINDOW_WIDTH/2, WINDOW_HEIGHT/2);
    units.push_front(midripod);
    units.push_front(cominex);

    cominex->energy->increase(100);
    midripod->energy->increase(100);
    midripod->powergem->increase(1);
    midripod->mine(grid);

    while(!stop) {
        // Update Model
        player_move(cominex, grid);
        for(struct unit * unit : units) {
            unit->cycle(&units, grid);
        }

        // Draw
        SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
        SDL_RenderClear(renderer);
        for(int x=0; x<GRID_WIDTH; x++) {
            for(int y=0; y<GRID_HEIGHT; y++) {
                grid[x][y].draw(renderer, x, y);
            }
        }
        for(struct unit * unit : units) {
            unit->draw(renderer);
        }
        building_menu->draw(renderer);
        SDL_RenderPresent(renderer);
        SDL_Delay(10);

        // Process Input
        while(SDL_PollEvent(&event)) {
            if(event.type == SDL_QUIT) {
                stop = 1;
            } else if(event.type == SDL_KEYDOWN) {
                if(event.key.keysym.sym == SDLK_ESCAPE) {
                    stop = 1;
                } else if(event.key.keysym.sym == SDLK_SPACE) {
                    cominex->mine(grid);
/*
                } else if(event.key.keysym.sym == SDLK_p) {
                    midripod->build(models["energen"], cominex->x, cominex->y, &units);
                } else if(event.key.keysym.sym == SDLK_s) {
                    midripod->build(models["enerstash"], cominex->x, cominex->y, &units);
                } else if(event.key.keysym.sym == SDLK_f) {
                    midripod->build(models["factory"], cominex->x, cominex->y, &units);
*/
                }
            } else if(event.type == SDL_MOUSEBUTTONDOWN) {
                struct model * model = building_menu->click(event.button.x, event.button.y);
                midripod->build(model, cominex->x, cominex->y, &units);
            }
        }
    }

    // Clean
    delete(building_menu);
    for(struct unit * unit : units) {
        delete(unit);
    }
/*
    for(struct model * model : models) {
        delete(model);
    }
*/
}


void player_move(struct unit * cominex, struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    const Uint8 * keyboard = SDL_GetKeyboardState(NULL);
    float dx = 0;
    float dy = 0;

    if(keyboard[SDL_SCANCODE_UP])
        dy--;
    if(keyboard[SDL_SCANCODE_DOWN])
        dy++;
    if(keyboard[SDL_SCANCODE_LEFT])
        dx--;
    if(keyboard[SDL_SCANCODE_RIGHT])
        dx++;
    if(dx && dy) {
        dx *= 0.7;
        dy *= 0.7;
    }
    cominex->move(dx, dy, grid);
}

void build_level(struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    cell matrix[GRID_WIDTH][GRID_HEIGHT];

    white_noise(matrix);
    // cell_automat_smooth(matrix, 3, 5, 3);
    cell_automat_smooth(matrix, 3, 3, 5);

    for(unsigned int x=0; x<GRID_WIDTH; x++) {
        for(unsigned int y=0; y<GRID_HEIGHT; y++) {
            if(matrix[x][y]) {
                if(rand()%20) { // Rock
                    grid[x][y] = square(HARDNESS_ROCK);
                } else {
                    if(rand()%4) { // Cristal
                        grid[x][y] = square(HARDNESS_CRISTAL);
                        grid[x][y].cristal = 2+rand()%3;
                    } else { // Power Gem
                        grid[x][y] = square(HARDNESS_POWERGEM);
                        grid[x][y].powergem = 1;
                    }
                }
            } else {
                grid[x][y] = square(0);
            }
        }
    }
}
